/* Bluetooth stack headers */
#include "bg_types.h"
#include "native_gecko.h"
#include "gatt_db.h"
#include "gpiointerrupt.h"

#include "app.h"
#define TIMER_ID		1
#define TIMER_TIME_ID	2

void setAdvScanData(void);

typedef struct
{
	uint8_t len_flags;
	uint8_t type_flags;
	uint8_t val_flags;

	uint8_t len_manuf;
	uint8_t type_manuf;
	uint8_t company_LO;
	uint8_t company_HI;
	char    manuf_name[12];
} customAdv_t;

typedef struct
{
	uint8_t len_name;
	uint8_t type_name;
	char    name[7];
} customScanResp_t;

static customAdv_t advData;
static customScanResp_t scanData;
static uint8_t counter, conn_handle;
static uint32_t duration;					// connection duration
static uint16_t notify_cccd, indicate_cccd;

/* Print boot message */
static void bootMessage(struct gecko_msg_system_boot_evt_t *bootevt);
static void Button_Event(const uint8_t pin);

/* Flag for indicating DFU Reset must be performed */
static uint8_t boot_to_dfu = 0;

/* Main application */
void appMain(gecko_configuration_t *pconfig)
{
#if DISABLE_SLEEP > 0
  pconfig->sleep.flags = 0;
#endif

  /* Initialize debug prints. Note: debug prints are off by default. See DEBUG_LEVEL in app.h */
  initLog();
  setAdvScanData();
  GPIO_PinModeSet(BSP_LED0_PORT, BSP_LED0_PIN, gpioModePushPull, 0);  // red LED
  GPIO_PinModeSet(BSP_LED1_PORT, BSP_LED1_PIN, gpioModePushPull, 0);  // green LED
  GPIO_PinModeSet(BSP_BUTTON0_PORT, BSP_BUTTON0_PIN, gpioModeInputPullFilter, 1); // button config
  GPIO_ExtIntConfig(BSP_BUTTON0_PORT, BSP_BUTTON0_PIN, BSP_BUTTON0_PIN, false, true, true);
  GPIOINT_Init();													// init GPIOINT driver
  GPIOINT_CallbackRegister(BSP_BUTTON0_PIN, Button_Event);   	// register callback function

  /* Initialize stack */
  gecko_init(pconfig);

  while (1) {
    /* Event pointer for handling events */
    struct gecko_cmd_packet* evt;

    /* if there are no events pending then the next call to gecko_wait_event() may cause
     * device go to deep sleep. Make sure that debug prints are flushed before going to sleep */
    if (!gecko_event_pending()) {
      flushLog();
    }

    /* Check for stack event. This is a blocking event listener. If you want non-blocking please see UG136. */
    evt = gecko_wait_event();

    /* Handle events */
    switch (BGLIB_MSG_ID(evt->header)) {
      /* This boot event is generated when the system boots up after reset.
       * Do not call any stack commands before receiving the boot event.
       * Here the system is set to start advertising immediately after boot procedure. */
      case gecko_evt_system_boot_id:
        bootMessage(&(evt->data.evt_system_boot));
        printLog("boot event - starting advertising\r\n");

        /* Set advertising parameters. 100ms advertisement interval.
         * The first parameter is advertising set handle
         * The next two parameters are minimum and maximum advertising interval, both in
         * units of (milliseconds * 1.6).
         * The last two parameters are duration and maxevents left as default. */
        gecko_cmd_le_gap_set_advertise_timing(0, 160, 160, 0, 0);
        gecko_cmd_system_set_tx_power(0);								// set max Tx power (0 dBm)
        gecko_cmd_le_gap_bt5_set_adv_data(0, 0, sizeof(advData), (const uint8_t*)&advData);	// set avd. data
        gecko_cmd_le_gap_bt5_set_adv_data(0, 1, sizeof(scanData), (const uint8_t*)&scanData);	// set scan response data
        /* Start general advertising and enable connections. */
        gecko_cmd_le_gap_start_advertising(0, le_gap_user_data, le_gap_connectable_scannable);
      break;

      case gecko_evt_le_connection_opened_id:
    	conn_handle = evt->data.evt_le_connection_opened.connection;
    	GPIO_PinOutSet(BSP_LED0_PORT, BSP_LED0_PIN);
    	gecko_cmd_hardware_set_soft_timer(32768, TIMER_ID, 0);
        printLog("connection opened\r\n");
        duration = 0;						// init connection duration
        gecko_cmd_gatt_server_write_attribute_value(gattdb_connect_time, 0,
            				  sizeof(duration), (uint8_t*)&duration);
      break;

      case gecko_evt_le_connection_closed_id:
    	gecko_cmd_hardware_set_soft_timer(0, TIMER_ID, 0);
    	GPIO_PinOutClear(BSP_LED0_PORT, BSP_LED0_PIN);
        printLog("connection closed, reason: 0x%2.2x\r\n", evt->data.evt_le_connection_closed.reason);
        /* Check if need to boot to OTA DFU mode */
        if (boot_to_dfu) {
          /* Enter to OTA DFU mode */
          gecko_cmd_system_reset(2);
        } else {
          /* Restart advertising after client has disconnected */
          gecko_cmd_le_gap_start_advertising(0, le_gap_general_discoverable, le_gap_connectable_scannable);
        }
      break;

      /* Events related to OTA upgrading
         ----------------------------------------------------------------------------- */
      /* Check if the user-type OTA Control Characteristic was written.
       * If ota_control was written, boot the device into Device Firmware Upgrade (DFU) mode. */
      case gecko_evt_gatt_server_user_write_request_id:
        if (evt->data.evt_gatt_server_user_write_request.characteristic == gattdb_ota_control) {
          /* Set flag to enter to OTA mode */
          boot_to_dfu = 1;
          /* Send response to Write Request */
          gecko_cmd_gatt_server_send_user_write_response(
            evt->data.evt_gatt_server_user_write_request.connection,
            gattdb_ota_control,
            bg_err_success);
          /* Close connection to enter to DFU OTA mode */
          gecko_cmd_le_connection_close(evt->data.evt_gatt_server_user_write_request.connection);
        }
        if (evt->data.evt_gatt_server_user_write_request.characteristic == gattdb_counter_value) {
        	counter = evt->data.evt_gatt_server_user_write_request.value.data[0];
        	gecko_cmd_gatt_server_send_user_write_response(conn_handle, gattdb_counter_value, 0);
        }
      break;

      /* Add additional event handlers as your application requires */
      case gecko_evt_gatt_server_user_read_request_id:  // client wants to read from DB
    	  if (evt->data.evt_gatt_server_attribute_value.attribute == gattdb_counter_value) {
    		  gecko_cmd_gatt_server_send_user_read_response(conn_handle, gattdb_counter_value, 0, 1, &counter);
    	  }
      break;

      case gecko_evt_hardware_soft_timer_id:
    	  if (evt->data.evt_hardware_soft_timer.handle == TIMER_ID) {
    		  counter++;
    		  if (notify_cccd) {
    			  gecko_cmd_gatt_server_send_characteristic_notification(0xFF,
    			  	  gattdb_counter_value, 1, (const uint8*)&counter);
    		  }
    		  // increment characteristic of type "hex" (read-modify-write)
    		  uint32_t *data = (uint32_t*)gecko_cmd_gatt_server_read_attribute_value(gattdb_connect_time, 0)->value.data;
    		  strncpy((char*)&duration, (char*)data, 4);  // copy characteristic from the Bluetooth DB
    		  duration++;	// update connection duration
    		  gecko_cmd_gatt_server_write_attribute_value(gattdb_connect_time, 0,
    			sizeof(duration), (uint8_t*)&duration);	  // write characteristic to the Bluetooth DB
    		  printLog("%02X\n\r", counter);
    	  }
    	  if (evt->data.evt_hardware_soft_timer.handle == TIMER_TIME_ID) {
    		  uint32_t *data = (uint32_t*)gecko_cmd_gatt_server_read_attribute_value(gattdb_connect_time,0)->value.data;
    		  uint32_t tmp;
    		  strncpy((char*)&tmp, (char*)data, 4);
    		  gecko_cmd_gatt_server_send_characteristic_notification(0xFF,
    		      			  	  gattdb_connect_time, 1, (const uint8*)&tmp);
    	  }
      break;

      case gecko_evt_gatt_server_characteristic_status_id:			// some attribute value changed by client
    	 if (evt->data.evt_gatt_server_characteristic_status.characteristic == gattdb_counter_value) {
    		if (evt->data.evt_gatt_server_characteristic_status.status_flags == gatt_server_client_config) { // CCCD is changed
    		    notify_cccd = evt->data.evt_gatt_server_characteristic_status.client_config_flags;
    		}
    		if (notify_cccd == 1)
    			printLog("Notifications enabled\n\r");
    		else
    			printLog("Notifications disabled\n\r");
    	 }
    	 if (evt->data.evt_gatt_server_characteristic_status.characteristic == gattdb_connect_time) {
    		 if (evt->data.evt_gatt_server_characteristic_status.status_flags == gatt_server_client_config) { // CCCD is changed
    		     uint16_t tmp = evt->data.evt_gatt_server_characteristic_status.client_config_flags;
    		     if (tmp == 1)
    		    	 gecko_cmd_hardware_set_soft_timer(32768*10, TIMER_TIME_ID, 0); // start notif timer
    		     else
    		    	 gecko_cmd_hardware_set_soft_timer(0, TIMER_TIME_ID, 0);		// stop notif timer
    		 }

    	 }
    	 if (evt->data.evt_gatt_server_characteristic_status.characteristic == gattdb_button_state) {
    		 if (evt->data.evt_gatt_server_characteristic_status.status_flags == gatt_server_confirmation) {
    			 printLog("Indicate ACK received\n\r");
    		 }
    		 else {
    			 indicate_cccd = evt->data.evt_gatt_server_characteristic_status.client_config_flags;
    			 if (indicate_cccd == 2)
    				 printLog("Indications enabled\n\r");
    			 else
    			     printLog("Indications disabled\n\r");
    		 }

    	 }
      break;

      case gecko_evt_system_external_signal_id:					// button event
    	  if (evt->data.evt_system_external_signal.extsignals == BSP_BUTTON0_PIN) {
    		  GPIO_PinOutToggle(BSP_LED1_PORT, BSP_LED1_PIN);	// toggle green LED
    		  uint8_t state = GPIO_PinInGet(BSP_LED1_PORT, BSP_LED1_PIN); // get button state
    		  gecko_cmd_gatt_server_write_attribute_value(gattdb_button_state, 0, 1, &state);
    		  if (indicate_cccd) {
    			  gecko_cmd_gatt_server_send_characteristic_notification(0xFF,
    					  gattdb_button_state, 1, (const uint8*)&state);
    		  }
    	  }
      break;

      default:
      break;
    }
  }
}

/* Print stack version and local Bluetooth address as boot message */
static void bootMessage(struct gecko_msg_system_boot_evt_t *bootevt)
{
#if DEBUG_LEVEL
  bd_addr local_addr;
  int i;

  printLog("stack version: %u.%u.%u\r\n", bootevt->major, bootevt->minor, bootevt->patch);
  local_addr = gecko_cmd_system_get_bt_address()->address;

  printLog("local BT device address: ");
  for (i = 0; i < 5; i++) {
    printLog("%2.2x:", local_addr.addr[5 - i]);
  }
  printLog("%2.2x\r\n", local_addr.addr[0]);
#endif
}

void setAdvScanData(void)
{
	advData.len_flags = 0x02;
	advData.type_flags = 0x01;
	advData.val_flags = 0x06;					// general discoverable mode

	advData.len_manuf = 15;
	advData.type_manuf = 0xFF;
	advData.company_LO = 0xFF;					// Silabs company ID (0x02FF)
	advData.company_HI = 0x02;
	strncpy(advData.manuf_name, "Silicon Labs", 12);

	scanData.len_name = 8;
	scanData.type_name = 9;						// full device name
	strncpy(scanData.name, "CSCI470", 7);
}

static void Button_Event(const uint8_t pin)
{
	if (!GPIO_PinInGet(BSP_BUTTON0_PORT, BSP_BUTTON0_PIN)) {
		gecko_external_signal(BSP_BUTTON0_PIN);	// inform Bluetooth stack on the button event
	}

}
